from flask import (
    current_app,
    render_template,
    request,
    jsonify,
    url_for
)
from six import string_types
from flask import Blueprint
from rqmonitor.utils import (
                            list_all_queues_names, list_all_possible_job_status,
                            list_all_queues, reformat_job_data, delete_workers,
                            create_redis_connection, delete_queue, empty_queue,
                            delete_job, cancel_job, requeue_job, get_redis_memory_used,
                            job_count_in_queue_registry, resolve_jobs,
                            delete_all_jobs_in_queues_registries, requeue_all_jobs_in_failed_registry,
                            cancel_all_queued_jobs
)

from rqmonitor.defaults import RQ_MONITOR_REFRESH_INTERVAL
from rq.connections import pop_connection, push_connection, get_current_connection
from rqmonitor.decorators import cache_control_no_store, catch_global_exception
from rqmonitor.exceptions import RQMonitorException
from rq.worker import Worker
from rq.suspension import suspend, resume, is_suspended
import logging
import socket


logger = logging.getLogger(__name__)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)
logger.setLevel(logging.INFO)

REDIS_RQ_HOST = 'localhost:6379'

monitor_blueprint = Blueprint('rqmonitor', __name__, template_folder='templates', static_folder='static')


@monitor_blueprint.errorhandler(RQMonitorException)
def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


@monitor_blueprint.before_app_first_request
def setup_redis_connection():
    redis_url = current_app.config.get("RQ_MONITOR_REDIS_URL")
    if isinstance(redis_url, string_types):
        # update as tuple
        current_app.config["RQ_MONITOR_REDIS_URL"] = (redis_url,)
        current_app.redis_connection = create_redis_connection((redis_url,)[0])
    elif isinstance(redis_url, (tuple, list)):
        current_app.redis_connection = create_redis_connection(redis_url[0])
    else:
        raise RuntimeError("No Redis configuration!")


@monitor_blueprint.before_request
def push_rq_connection():
    new_instance_index = None
    if request.method == 'GET':
        new_instance_index = request.args.get('redis_instance_index')
    if request.method == 'POST':
        new_instance_index = request.form.get('redis_instance_index')

    if new_instance_index is None:
        new_instance_index = request.view_args.get('redis_instance_index')

    #print(new_instance_index, type(new_instance_index))
    if new_instance_index is not None:
        redis_url = current_app.config.get("RQ_MONITOR_REDIS_URL")
        new_instance_index = int(new_instance_index)
        if int(new_instance_index) < len(redis_url):
            new_instance = create_redis_connection(redis_url[new_instance_index])
        else:
            raise RQMonitorException("Invalid redis instance index!", status_code=400)
    else:
        new_instance = current_app.redis_connection
    push_connection(new_instance)
    current_app.redis_connection = new_instance


@monitor_blueprint.teardown_request
def pop_rq_connection(exception=None):
    pop_connection()


@monitor_blueprint.route('/', defaults={"redis_instance_index": 0})
@catch_global_exception
@cache_control_no_store
def home(redis_instance_index):
    rq_queues_list = list_all_queues_names()
    rq_possible_job_status = list_all_possible_job_status()
    site_map = {}
    for rule in current_app.url_map.iter_rules():
        if "static" not in rule.endpoint:
            site_map[rule.endpoint] = url_for(rule.endpoint)

    return render_template('rqmonitor/index.html',
                                rq_queues_list= rq_queues_list,
                                rq_possible_job_status= rq_possible_job_status,
                                redis_instance_list=current_app.config.get('RQ_MONITOR_REDIS_URL'),
                                redis_memory_used=get_redis_memory_used(),
                                site_map=site_map
                           )


@monitor_blueprint.route('/jobs_dashboard')
@catch_global_exception
@cache_control_no_store
def get_jobs_dashboard():
    return render_template('rqmonitor/jobs.html')


@monitor_blueprint.route('/workers_dashboard')
@catch_global_exception
@cache_control_no_store
def get_workers_dashboard():
    return render_template('rqmonitor/workers.html',
                            is_suspended=is_suspended(connection=get_current_connection()))


@monitor_blueprint.route('/queues_dashboard')
@catch_global_exception
@cache_control_no_store
def get_queues_dashboard():
    return render_template('rqmonitor/queues.html')


@monitor_blueprint.route('/queues', methods=['GET'])
@catch_global_exception
@cache_control_no_store
def list_queues_api():
    queue_list = list_all_queues()
    rq_queues = []
    for queue in queue_list:
        rq_queues.append({
                'queue_name': queue.name,
                'job_count': queue.count,
            }
        )
    return {
        'data': rq_queues,
    }


@monitor_blueprint.route('/workers', methods=['GET'])
@catch_global_exception
@cache_control_no_store
def list_workers_api():
    workers_list = Worker.all()
    rq_workers = []
    for worker in workers_list:
        host_ip_using_name = "Unresolved"
        try:
            host_ip_using_name = socket.gethostbyname(worker.hostname)
        except socket.gaierror as addr_error:
            pass

        rq_workers.append(
            {
                'worker_name': worker.name,
                'listening_on': ', '.join(queue.name for queue in worker.queues),
                'status': worker.get_state() if not is_suspended(get_current_connection()) else "suspended",
                'host_ip': host_ip_using_name,
                'current_job_id': worker.get_current_job_id(),
                'failed_jobs': worker.failed_job_count,
            }
        )
    return {
        'data': rq_workers,
    }


@monitor_blueprint.route('queues/sidebar', methods=['GET'])
@catch_global_exception
@cache_control_no_store
def refresh_sidebar_queues():
    return render_template('rqmonitor/sidebar_queues.html',
        rq_queues_list=list_all_queues_names()
    )


@monitor_blueprint.route('/jobs', methods=['GET'])
@catch_global_exception
@cache_control_no_store
def list_jobs_api():
    """
    :param request: Flask GET request containing two parameters acting as filter for jobs
                    1) Jobs Status list (with these status)
                    2) queues list (to fetch queues)
    :return: rendered output
    """
    serialised_jobs = []

    start = int(request.args.get('start'))
    length = int(request.args.get('length'))
    draw = int(request.args.get('draw'))
    search = request.args.get('search[value]')
    requested_queues = request.args.getlist('queues[]')
    requested_job_status = request.args.getlist('jobstatus[]')

    if not requested_queues or not requested_job_status:
        return {
            'draw': draw,
            'recordsTotal': 0,
            'recordsFiltered': 0,
            'data': serialised_jobs,
        }

    job_counts = []
    total_job_count = 0

    for queue in requested_queues:
        for job_status in requested_job_status:
            queue_registry_count = job_count_in_queue_registry(queue, job_status)
            job_counts.append((queue, job_status, queue_registry_count))
            total_job_count += queue_registry_count

    jobs = resolve_jobs(job_counts, start, length)

    for job in jobs:
        serialised_jobs.append(reformat_job_data(job))

    return {
        'draw': draw,
        'recordsTotal': total_job_count,
        'recordsFiltered': total_job_count,
        'data': serialised_jobs,
    }


@monitor_blueprint.route('/workers/delete', methods=['POST'])
@cache_control_no_store
def delete_workers_api():
    worker_names = []
    if request.method == 'POST':
        worker_id = request.form.get('worker_id', None)
        delete_all = request.form.get('delete_all')
        if worker_id == None and (delete_all == "false" or delete_all == None):
            raise RQMonitorException('Worker ID not received', status_code=400)

        if delete_all == "true":
            for worker_instance in Worker.all():
                worker_names.append(worker_instance.name)
            delete_workers(worker_names)
        else:
            worker_names.append(worker_id)
            delete_workers([worker_id])

        return {
            'message': 'Successfully deleted worker/s {0}'.format(", ".join(worker_names))
        }


@monitor_blueprint.route('/workers/suspend', methods=['POST'])
@cache_control_no_store
def suspend_workers_api():
    if request.method == 'POST':
        suspend(connection=get_current_connection())
        return {
            'message': 'Successfully suspended all workers'
        }


@monitor_blueprint.route('/workers/resume', methods=['POST'])
@catch_global_exception
@cache_control_no_store
def resume_workers_api():
    if request.method == 'POST':
        resume(connection=get_current_connection())
        return {
            'message': 'Successfully resumed all workers'
        }


@monitor_blueprint.route('/queues/delete', methods=['POST'])
@catch_global_exception
@cache_control_no_store
def delete_queue_api():
    if request.method == 'POST':
        queue_id = request.form.get('queue_id', None)
        if queue_id is None:
            raise RQMonitorException('Queue Name not received',  status_code=400)
        delete_queue(queue_id)
        return {
            'message': 'Successfully deleted {0}'.format(queue_id)
        }


@monitor_blueprint.route('/queues/empty', methods=['POST'])
@catch_global_exception
@cache_control_no_store
def empty_queue_api():
    if request.method == 'POST':
        queue_id = request.form.get('queue_id', None)
        if queue_id is None:
            raise RQMonitorException('Queue Name not received', status_code=400)

        empty_queue(queue_id)

        return {
            'message': 'Successfully emptied {0}'.format(queue_id)
        }


@monitor_blueprint.route('/queues/delete/all', methods=['POST'])
@cache_control_no_store
def delete_all_queues_api():
    if request.method == 'POST':
        queue_names = [queue.name for queue in list_all_queues()]
        for queue in list_all_queues():
            queue.delete(delete_jobs=True)
        return {
            'message': 'Successfully deleted queues {0}'.format(", ".join(queue_names))
        }


@monitor_blueprint.route('/queues/empty/all', methods=['POST'])
@catch_global_exception
@cache_control_no_store
def empty_all_queues_api():
    if request.method == 'POST':
        queue_names = [queue.name for queue in list_all_queues()]
        for queue in list_all_queues():
            queue.empty()
        return {
            'message': 'Successfully emptied queues {0}'.format(", ".join(queue_names))
        }


@monitor_blueprint.route('/workers/info', methods=['GET'])
@catch_global_exception
@cache_control_no_store
def worker_info_api():
    if request.method == 'GET':
        worker_id = request.args.get('worker_id', None)

        if worker_id is None:
            raise RQMonitorException('Worker ID not received !', status_code=400)

        worker_instance = Worker.find_by_key(Worker.redis_worker_namespace_prefix + worker_id)
        return {
            'worker_host_name': worker_instance.hostname.decode('utf-8'),
            'worker_ttl': worker_instance.default_worker_ttl,
            'worker_result_ttl': worker_instance.default_result_ttl,
            'worker_name': worker_instance.name,
            'worker_birth_date': worker_instance.birth_date.strftime('%d-%m-%Y %H:%M:%S')
                                if worker_instance.birth_date is not None else "Not Available",
            'worker_death_date': worker_instance.death_date.strftime('%d-%m-%Y %H:%M:%S')
                                if worker_instance.death_date is not None else "Is Alive",
            'worker_last_cleaned_at': worker_instance.last_cleaned_at.strftime('%d-%m-%Y %H:%M:%S')
                                if worker_instance.last_cleaned_at is not None else "Not Yet Cleaned",
            'worker_failed_job_count': worker_instance.failed_job_count,
            'worker_successful_job_count': worker_instance.successful_job_count,
            'worker_job_monitoring_interval': worker_instance.job_monitoring_interval,
            'worker_last_heartbeat': worker_instance.last_heartbeat.strftime('%d-%m-%Y %H:%M:%S')
                                if worker_instance.last_heartbeat is not None else "Not Available",
            'worker_current_job_id': worker_instance.get_current_job_id(),
        }


@monitor_blueprint.route('/jobs/cancel', methods=['POST'])
@catch_global_exception
@cache_control_no_store
def cancel_job_api():
    if request.method == 'POST':
        job_id = request.form.get('job_id')
        if job_id is None:
            raise RQMonitorException('Job ID not received', status_code=400)

        cancel_job(job_id)

        return {
            'message': 'Successfully cancelled job with ID {0}'.format(job_id)
        }


@monitor_blueprint.route('/jobs/requeue',  methods=['POST'])
@catch_global_exception
@cache_control_no_store
def requeue_job_api():
    if request.method == 'POST':
        job_id = request.form.get('job_id')
        if job_id is None:
            raise RQMonitorException('Job ID not received', status_code=400)

        requeue_job(job_id)

        return {
            'message': 'Successfully requeued job with ID {}'.format(job_id)
        }


@monitor_blueprint.route('/jobs/delete',  methods=['POST'])
@catch_global_exception
@cache_control_no_store
def delete_job_api():
    if request.method == 'POST':
        job_id = request.form.get('job_id')
        if job_id is None:
            raise RQMonitorException('Job ID not received', status_code=400)

        delete_job(job_id)
        return {
            'message': 'Successfully deleted job with ID {0}'.format(job_id)
        }


@monitor_blueprint.route('/jobs/delete/all',  methods=['POST'])
@catch_global_exception
@cache_control_no_store
def delete_all_jobs_api():
    if request.method == 'POST':
        requested_queues = request.form.getlist('queues[]')
        requested_job_status = request.form.getlist('jobstatus[]')

        if requested_queues is None or requested_job_status is None:
            raise RQMonitorException('No queue/status selected', status_code=400)

        delete_all_jobs_in_queues_registries(requested_queues, requested_job_status)

        return {
            'message': 'Successfully deleted all jobs with status as {0} on queues {1}'
                .format(", ".join(requested_job_status), ", ".join(requested_queues))
        }


@monitor_blueprint.route('/jobs/requeue/all',  methods=['POST'])
@catch_global_exception
@cache_control_no_store
def requeue_failed_jobs_api():
    if request.method == 'POST':
        requested_queues = request.form.getlist('queues[]')
        if requested_queues is None:
            raise RQMonitorException('No queue/s selected', status_code=400)

        fail_count = requeue_all_jobs_in_failed_registry(requested_queues)

        return {
            'message': 'Successfully requeued all jobs on queues {0}'.format(", ".join(requested_queues))
        }


@monitor_blueprint.route('/jobs/cancel/all',  methods=['POST'])
@catch_global_exception
@cache_control_no_store
def cancel_queued_jobs_api():
    if request.method == 'POST':
        requested_queues = request.form.getlist('queues[]')
        if requested_queues is None:
            raise RQMonitorException('No queue/s selected', status_code=400)

        fail_count = cancel_all_queued_jobs(requested_queues)

        return {
            'message': 'Successfully requeued all jobs'
        }


@monitor_blueprint.route('/redis/memory')
@catch_global_exception
@cache_control_no_store
def redis_memory_api():
    return {
        'redis_memory_used': get_redis_memory_used()
    }


@monitor_blueprint.context_processor
def inject_refresh_interval():
    refresh_interval = current_app.config.get("RQ_MONITOR_REFRESH_INTERVAL",
                                              RQ_MONITOR_REFRESH_INTERVAL)
    return dict(refresh_interval=refresh_interval)


@monitor_blueprint.context_processor
def inject_redis_instances_count():
    return dict(redis_instances_count=len(current_app.config["RQ_MONITOR_REDIS_URL"]))


import traceback


class RQMonitorException(Exception):
    '''
    used to catch any possible exception that might occur in used libraries
    or while performing some action and display gracefully inside rqmonitor
    '''
    def __init__(self, message='RQ Monitor Global Exception',
                 status_code=500, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        rv['traceback'] = ''.join(traceback.format_tb(self.__traceback__))
        return rv



import sys
import traceback
from functools import wraps
from flask import make_response
from rqmonitor.exceptions import RQMonitorException


def cache_control_no_store(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        _rendered_template = func(*args, **kwargs)
        _make_response = make_response(_rendered_template)
        _make_response.headers.set("Cache-Control", "no-store")
        return _make_response
    return _wrapper


def catch_global_exception(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            inner_response = func(*args, **kwargs)
        except Exception as e:
            tb = sys.exc_info()[2]
            error_message = getattr(e, 'message', None)
            status_code = getattr(e, 'status_code', None)
            kwargs = {}
            if error_message is not None:
                kwargs.update({'message': error_message})
            if status_code is not None:
                kwargs.update({'status_code': status_code})
            raise RQMonitorException(**kwargs).with_traceback(tb)
        return inner_response
    return _wrapper


import humanize
import redis
import os
import signal
import logging
import socket
import zlib
import errno
from rq.registry import (StartedJobRegistry,
                         FinishedJobRegistry,
                         FailedJobRegistry,
                         DeferredJobRegistry,
                         ScheduledJobRegistry)
from rq.connections import resolve_connection
from rq.utils import utcparse
from rq.exceptions import InvalidJobOperationError
from rqmonitor.exceptions import RQMonitorException
from datetime import datetime
from rq.worker import Worker
from rq.queue import Queue
from rq.job import Job
from fabric import Connection, Config
from invoke import UnexpectedExit


logger = logging.getLogger(__name__)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)
logger.setLevel(logging.INFO)



REGISTRIES = [StartedJobRegistry, FinishedJobRegistry,
              FailedJobRegistry, DeferredJobRegistry,
              ScheduledJobRegistry]

JobStatus = ['queued', 'finished', 'failed', 'started', 'deferred', 'scheduled']


def create_redis_connection(redis_url):
    return redis.Redis.from_url(redis_url)

def send_signal_worker(worker_id):
    worker_instance = Worker.find_by_key(Worker.redis_worker_namespace_prefix+worker_id)
    worker_instance.request_stop(signum=2, frame=5)


def delete_queue(queue_id):
    """
    :param queue_id: Queue ID/name to delete
    :return: None

    As no specific exception is raised for below method
    we are using general Exception class for now
    """
    def attach_rq_queue_prefix(queue_id):
        return Queue.redis_queue_namespace_prefix + queue_id

    queue_instance = Queue.from_queue_key(attach_rq_queue_prefix(queue_id))
    queue_instance.delete()


def empty_queue(queue_id):
    """
    :param queue_id: Queue ID/name to delete
    :return: None

    As no specific exception is raised for below method
    we are using general Exception class for now
    """
    def attach_rq_queue_prefix(queue_id):
        return Queue.redis_queue_namespace_prefix + queue_id

    queue_instance = Queue.from_queue_key(attach_rq_queue_prefix(queue_id))
    queue_instance.empty()


def delete_workers(worker_ids, signal_to_pass=signal.SIGINT):
    """
    Expect worker ID without RQ REDIS WORKER NAMESPACE PREFIX of rq:worker:
    By default performs warm shutdown

    :param worker_id: list of worker id's to delete
    :param signal_to_pass:
    :return:
    """
    # find worker instance by key, refreshes worker implicitly
    def attach_rq_worker_prefix(worker_id):
        return Worker.redis_worker_namespace_prefix + worker_id

    for worker_instance in [Worker.find_by_key(attach_rq_worker_prefix(worker_id))
                            for worker_id in worker_ids]:
        requested_hostname = worker_instance.hostname
        requested_hostname = requested_hostname.decode('utf-8')
        # kill if on same instance
        if socket.gethostname() == requested_hostname:
            os.kill(worker_instance.pid, signal_to_pass)
        else:
            required_host_ip = socket.gethostbyname(requested_hostname)
            fabric_config_wrapper = Config()
            # loads from user level ssh config (~/.ssh/config) and system level
            # config /etc/ssh/ssh_config
            fabric_config_wrapper.load_ssh_config()
            # to use its ssh_config parser abilities
            paramiko_ssh_config = fabric_config_wrapper.base_ssh_config
            for hostname in paramiko_ssh_config.get_hostnames():
                ssh_info = paramiko_ssh_config.lookup(hostname)
                available_host_ip = ssh_info.get('hostname')
                if available_host_ip == required_host_ip:
                    process_owner = None
                    # make connection via fabric and send SIGINT for now
                    ssh_connection = Connection(hostname)
                    try:
                        #find owner of process https://unix.stackexchange.com/questions/284934/return-owner-of-process-given-pid
                        process_owner = ssh_connection.run('ps -o user= -p {0}'.format(worker_instance.pid))
                        # have permission to kill so this works without sudo
                        # need to plan for other cases
                        process_owner = process_owner.stdout.strip(' \n\t')
                        result_kill = ssh_connection.run('kill -{0} {1}'.format(2, worker_instance.pid), hide=True)
                        if result_kill.failed:
                            raise RQMonitorException("Some issue occured on running command {0.command!r} "
                                                     "on {0.connection.host}, we got stdout:\n{0.stdout}"
                                                     "and stderr:\n{0.stderr}".format(result_kill))
                    except UnexpectedExit as e:
                        stdout, stderr = e.streams_for_display()
                        # plan to accept password from user and proceed with sudo in future
                        if "Operation not permitted" in stderr.strip(' \n\t'):
                            raise RQMonitorException('Logged in user {0} does not have permission to kill worker'
                                                     ' process with pid {1} on {2} because it is owned '
                                                     ' by user {3}'.format(ssh_info.get('user'), worker_instance.pid,
                                                                           required_host_ip, process_owner))
                        raise RQMonitorException('Invoke\'s UnexpectedExit occurred with'
                                                 'stdout: {0}\nstderr: {1}\nresult: {2}\nreason {3}'.format(stdout.strip(' \n\t'),
                                                                                                            stderr.strip(' \n\t'),
                                                                                                            e.result, e.reason))
                    return


def list_all_queues():
    """
    :return: Iterable for all available queue instances
    """
    return Queue.all()


def list_all_possible_job_status():
    """
    :return: list of all possible job status
    """
    return JobStatus


def list_all_queues_names():
    """
    :return: Iterable of all queue names
    """
    return [queue.name for queue in list_all_queues()]


# a bit hacky for now
def validate_job_data(val, default="None", humanize_func=None,
                      with_utcparse=False, relative_to_now=False, append_s=False):
    if not val:
        return default

    if humanize_func == None and append_s == True:
        return str(val)+"s"
    elif humanize_func == None:
        return val
    else:
        if with_utcparse and relative_to_now:
            return humanize_func(utcparse(val).timestamp() - datetime.now().timestamp())
        elif with_utcparse and not relative_to_now:
            return humanize_func(utcparse(val).timestamp())
        else:
            return humanize_func(val)


def reformat_job_data(job: Job):
    """
    Create serialized version of Job which can be consumed by DataTable
    (RQ provides to_dict) including origin(queue), created_at, data, description,
    enqueued_at, started_at, ended_at, result, exc_info, timeout, result_ttl,
     failure_ttl, status, dependency_id, meta, ttl

    :param job: Job Instance need to be serialized
    :return: serialized job
    """
    serialized_job = job.to_dict()
    return {
        "job_info": {
            "job_id": validate_job_data(job.get_id()),
            "job_description": validate_job_data(serialized_job.get('description')),
            "job_exc_info": validate_job_data(zlib.decompress(serialized_job.get('exc_info')).decode('utf-8')
                                              if serialized_job.get('exc_info') is not None
                                              else None),
            "job_status": validate_job_data(serialized_job.get('status')),
            "job_queue": validate_job_data(serialized_job.get('origin')),
            "job_created_time_humanize": validate_job_data(serialized_job.get('created_at'),
                                                           humanize_func=humanize.naturaltime,
                                                           with_utcparse=True,
                                                           relative_to_now=True),
            "job_enqueued_time_humanize": validate_job_data(serialized_job.get('enqueued_at'),
                                                           humanize_func=humanize.naturaltime,
                                                           with_utcparse=True,
                                                           relative_to_now=True),
            "job_ttl": validate_job_data(serialized_job.get('ttl'), default='Infinite', append_s=True),
            "job_timeout": validate_job_data(serialized_job.get('timeout'), default='180s', append_s=True),
            "job_result_ttl": validate_job_data(serialized_job.get('result_ttl'), default='500s', append_s=True),
            "job_fail_ttl": validate_job_data(serialized_job.get('failure_ttl'), default='1yr', append_s=True),
        },
    }


def get_queue(queue):
    """
    :param queue: Queue Name or Queue ID or Queue Redis Key or Queue Instance
    :return: Queue instance
    """
    if isinstance(queue, Queue):
        return queue

    if isinstance(queue, str):
        if queue.startswith(Queue.redis_queue_namespace_prefix):
            return Queue.from_queue_key(queue)
        else:
            return Queue.from_queue_key(Queue.redis_queue_namespace_prefix+queue)

    raise TypeError('{0} is not of class {1} or {2}'.format(queue, str, Queue))


def list_jobs_on_queue(queue):
    """
    If no worker has started jobs are not available in registries
    Worker does movement of jobs across registries
    :param queue: Queue to fetch jobs from
    :return: all valid jobs untouched by workers
    """
    queue = get_queue(queue)
    return queue.jobs


def list_job_ids_on_queue(queue):
    """
    If no worker has started jobs are not available in registries
    Worker does movement of jobs across registries
    :param queue: Queue to fetch jobs from
    :return: all valid jobs untouched by workers
    """
    queue = get_queue(queue)
    return queue.job_ids


def list_jobs_in_queue_all_registries(queue):
    """
    :param queue: List all jobs in all registries of given queue
    :return: list of jobs
    """
    jobs = []
    for registry in REGISTRIES:
        jobs.extend(list_jobs_in_queue_registry(queue, registry))
    return jobs


def list_jobs_in_queue_registry(queue, registry, start, end):
    """
    :param queue: Queue name from which jobs need to be listed
    :param registry: registry class from which jobs to be returned, default is all registries
    :return: list of all jobs matching above criteria
    """
    queue = get_queue(queue)
    if registry == StartedJobRegistry or registry == 'started':
        job_ids = queue.started_job_registry.get_job_ids(start, end)
        return [Job.fetch(job_id) for job_id in job_ids]
    elif registry == FinishedJobRegistry or registry == 'finished':
        job_ids = queue.finished_job_registry.get_job_ids(start, end)
        return [Job.fetch(job_id) for job_id in job_ids]
    elif registry == FailedJobRegistry or registry == 'failed':
        job_ids = queue.failed_job_registry.get_job_ids(start, end)
        return [Job.fetch(job_id) for job_id in job_ids]
    elif registry == DeferredJobRegistry or registry == 'deferred':
        job_ids = queue.deferred_job_registry.get_job_ids(start, end)
        return [Job.fetch(job_id) for job_id in job_ids]
    elif registry == ScheduledJobRegistry or registry == 'scheduled':
        job_ids = queue.scheduled_job_registry.get_job_ids(start, end)
        return [Job.fetch(job_id) for job_id in job_ids]
    # although not implemented as registry this is for ease
    elif registry == 'queued':
        return queue.get_jobs(start, end-start+1)
    else:
        return []


def list_job_ids_in_queue_registry(queue, registry, start=0, end=-1):
    """
    :param queue: Queue name from which jobs need to be listed
    :param registry: registry class from which jobs to be returned, default is all registries
    :return: list of all jobs matching above criteria
    """
    queue = get_queue(queue)
    if registry == StartedJobRegistry or registry == 'started':
        return queue.started_job_registry.get_job_ids(start, end)
    elif registry == FinishedJobRegistry or registry == 'finished':
        return queue.finished_job_registry.get_job_ids(start, end)
    elif registry == FailedJobRegistry or registry == 'failed':
        return queue.failed_job_registry.get_job_ids(start, end)
    elif registry == DeferredJobRegistry or registry == 'deferred':
        return queue.deferred_job_registry.get_job_ids(start, end)
    elif registry == ScheduledJobRegistry or registry == 'scheduled':
        return queue.scheduled_job_registry.get_job_ids(start, end)
    # although not implemented as registry this is for ease and uniformity
    elif registry == 'queued':
        return queue.get_job_ids(start, end-start+1)
    else:
        return []


def empty_registry(registry_name, queue_name, connection=None):
    """Empties a specific registry for a specific queue, Not in RQ, implemented
            here for performance reasons
    """
    redis_connection = resolve_connection(connection)
    queue_instance = Queue.from_queue_key(Queue.redis_queue_namespace_prefix+queue_name)

    registry_instance = None
    if registry_name == 'failed':
        registry_instance = queue_instance.failed_job_registry
    elif registry_name == 'started':
        registry_instance = queue_instance.started_job_registry
    elif registry_name == 'scheduled':
        registry_instance = queue_instance.scheduled_job_registry
    elif registry_name == 'deferred':
        registry_instance = queue_instance.deferred_job_registry
    elif registry_name == 'finished':
        registry_instance = queue_instance.finished_job_registry

    script = """
        local prefix = "{0}"
        local q = KEYS[1]
        local count = 0
        while true do
            local job_id, score = unpack(redis.call("zpopmin", q))
            if job_id == nil or score == nil then
                break
            end

            -- Delete the relevant keys
            redis.call("del", prefix..job_id)
            redis.call("del", prefix..job_id..":dependents")
            count = count + 1
        end
        return count
    """.format(registry_instance.job_class.redis_job_namespace_prefix).encode("utf-8")
    script = redis_connection.register_script(script)
    return script(keys=[registry_instance.key])


def delete_all_jobs_in_queues_registries(queues, registries):
    for queue in queues:
        for registry in registries:
            if registry == 'queued':
                # removes all jobs from queue and from job namespace
                get_queue(queue).empty()
            else:
                empty_registry(registry, queue)


def requeue_all_jobs_in_failed_registry(queues):
    fail_count = 0
    for queue in queues:
        failed_job_registry = get_queue(queue).failed_job_registry
        job_ids = failed_job_registry.get_job_ids()

        for job_id in job_ids:
            try:
                failed_job_registry.requeue(job_id)
            except InvalidJobOperationError:
                fail_count += 1

    return fail_count


def cancel_all_queued_jobs(queues):
    for queue in queues:
        job_ids = get_queue(queue).get_job_ids()
        for job_id in job_ids:
            Job.fetch(job_id).cancel()


def job_count_in_queue_registry(queue, registry):
    """
    :param queue: Queue name from which jobs need to be listed
    :param registry: registry class from which jobs to be returned, default is all registries
    :return: list of all jobs matching above criteria
    """
    queue = get_queue(queue)
    if registry == StartedJobRegistry or registry == 'started':
        return len(queue.started_job_registry)
    elif registry == FinishedJobRegistry or registry == 'finished':
        return len(queue.finished_job_registry)
    elif registry == FailedJobRegistry or registry == 'failed':
        return len(queue.failed_job_registry)
    elif registry == DeferredJobRegistry or registry == 'deferred':
        return len(queue.deferred_job_registry)
    elif registry == ScheduledJobRegistry or registry == 'scheduled':
        return len(queue.scheduled_job_registry)
    # although not implemented as registry this is for ease
    elif registry == 'queued':
        return len(queue)
    else:
        return 0


def get_redis_memory_used(connection=None):
    """
    All memory used in redis rq: namespace
    :param connection:
    :return:
    """
    RQ_REDIS_NAMESPACE = 'rq:*'
    redis_connection = resolve_connection(connection)
    script = """
        local sum = 0;
        local keys = {};
        local done = false;
        local cursor = "0"
        repeat
            local result = redis.call("SCAN", cursor, "match", ARGV[1])
            cursor = result[1];
            keys = result[2];
            for i, key in ipairs(keys) do
                local mem = redis.call("MEMORY", "USAGE", key);
                sum = sum + mem;
            end
            if cursor == "0" then
                done = true;
            end
        until done
        return sum;
    """
    script = redis_connection.register_script(script)
    return humanize.naturalsize(script(args=[RQ_REDIS_NAMESPACE]))


def delete_job(job_id):
    """
    Deleted job from the queue
    Does a implicit cancel with Job hash deleted
    :param job_id: Job id to be deleted
    :return: None
    """
    job_instance = Job.fetch(job_id)
    job_instance.delete(remove_from_queue=True)


def requeue_job(job_id):
    """
    Requeue job from the queue
    Will work only if job was failed
    :param job_id: Job id to be requeue
    :return: None

    """
    job_instance = Job.fetch(job_id)
    job_instance.requeue()


def cancel_job(job_id):
    """
    Only removes job from the queue
    :param job_id: Job id to be cancelled
    :return: None
    """
    job_instance = Job.fetch(job_id)
    job_instance.cancel()


def resolve_jobs(job_counts, start, length):

    start_block = 0
    cursor = 0
    jobs = []

    cumulative_count = 0
    for i, job_count in enumerate(job_counts):
        cumulative_count += job_count[2]
        if cumulative_count > start:
            start_block = i
            cursor = start - (cumulative_count - job_count[2])
            break

    for i, block in enumerate(job_counts[start_block:]):
        current_block_length = block[2]
        while cursor < current_block_length:
            current_block_jobs = list_jobs_in_queue_registry(block[0], block[1], cursor,
                                                             cursor+length-1)
            cursor += length
            jobs.extend(current_block_jobs)
            if len(jobs) >= length:
                return jobs[:length]
        cursor = 0

    return jobs[:length]



import os

from setuptools import find_packages, setup


def get_version():
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, 'rqmonitor/version.py')) as f:
        locals = {}
        exec(f.read(), locals)
        return locals['VERSION']


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='rqmonitor',
    version=get_version(),
    author="Pranav Gupta",
    author_email="pranavgupta4321@gmail.com",
    description="Flask based dynamic and actionable dashboard for monitoring RQs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pranavgupta1234/rqmonitor",
    download_url="https://github.com/pranavgupta1234/rqmonitor/archive/v_0.0.1.tar.gz",
    license="Apache Software License",
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'redis>=3.3.11',
        'humanize>=2.4.0',
        'Flask>=1.1.1',
        'Click>=7.0',
        'six>=1.13.0',
        'Werkzeug>=0.16.0',
        'rq>=1.2.0',
        'fabric>=2.5.0',
        'invoke>=1.4.1',
    ],
    entry_points={
        'console_scripts': [
            'rqmonitor = rqmonitor.cli:main'
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
 )
